# frozen_string_literal: true

require 'json'

module Dynamo
  module Record
    module BatchGet
      BATCH_SIZE = 100

      module ClassMethods
        def batch_get(keys)
          process_batch_get_requests(keys)
        end

        private

        def process_batch_get_requests(keys) # rubocop:disable Metrics/AbcSize, Metrics/CyclomaticComplexity
          attempt = 1
          results = []
          unprocessed_keys = keys.dup.map(&:stringify_keys)
          while !unprocessed_keys.empty? && attempt <= max_retries
            batch = unprocessed_keys.shift(BATCH_SIZE)
            Dynamo::Record.logger.debug "#{name} batch_get_item #{batch.count} request_items"
            processed, unprocessed = get_batch_from_dynamo(batch)
            results.concat(processed)

            next if unprocessed.empty?

            Dynamo::Record.logger.debug "#{name} batch_get_item #{unprocessed.count} unprocessed_keys"
            unprocessed_keys.unshift(*unprocessed)
            sleep(rand(1 << attempt) + 1)
            attempt += 1
          end

          if !unprocessed_keys.empty? && attempt > max_retries
            raise NumberOfRetriesExceeded.new(keys), 'Number of retries exceeded'
          end

          results.map { |r| build_item_from_hash(r) }
        end

        # BatchGetItem will return a partial result if provisioned throughput is
        # exceeded or an internal error occurs. If all items cannot be processed
        # ProvisionedThroughputExceededException is raised.
        # https://docs.aws.amazon.com/sdkforruby/api/Aws/DynamoDB/Client.html#batch_get_item-instance_method
        def get_batch_from_dynamo(batch)
          response = dynamodb_client.batch_get_item(
            request_items: {
              table_name => { keys: batch }
            }
          )
          [
            response.responses[table_name],
            response.unprocessed_keys[table_name]&.keys || []
          ]
        rescue Aws::DynamoDB::Errors::ProvisionedThroughputExceededException
          [[], batch]
        end

        # Convert response hashes to Record instances
        # Adapted from `build_item_from_resp` used by `find` but that method is closely tied to get_item response format
        # https://github.com/aws/aws-sdk-ruby-record/blob/v2.3.0/lib/aws-record/record/item_operations.rb#L593
        def build_item_from_hash(resp)
          record = new
          data = record.instance_variable_get('@data')
          attributes.attributes.each do |name, attr|
            data.set_attribute(name, attr.extract(resp))
            data.new_record = false
          end
          record
        end
      end

      class NumberOfRetriesExceeded < RuntimeError
        attr_reader :keys

        def initialize(keys)
          super
          @keys = keys
        end
      end
    end
  end
end
