# frozen_string_literal: true

module Dynamo
  module Record
    module Model
      COMPOSITE_DELIMITER = '|'
      DEFAULT_MAX_RETRIES = 5

      def self.included(klass)
        klass.include(Aws::Record)
        klass.include(Dynamo::Record::Marshalers)

        klass.extend ClassMethods
        klass.extend Dynamo::Record::BatchWrite::ClassMethods
        klass.extend Dynamo::Record::BatchGet::ClassMethods
        klass.send :prepend, InstanceMethods
      end

      module ClassMethods
        def table_name
          [Rails.configuration.dynamo['prefix'], name.tableize].join('-').tr('/', '.')
        end

        def scan
          raise 'no scanning in production' if Rails.env.production?

          super
        end

        def find(opts)
          super(opts).tap do |record|
            unless record
              name = self.name.demodulize
              conditions = opts.map { |k, v| "#{k}=#{v}" }.join(', ')
              error = "Couldn't find #{name} with #{conditions}"
              raise Aws::Record::Errors::NotFound, error
            end
          end
        end

        def find_all_by_hash_key(hash_key_value, opts = {})
          find_all_by_index_and_hash_key(hash_key, hash_key_value, opts)
        end

        def find_all_by_gsi_hash_key(gsi_name, hash_key_value, opts = {})
          hash_key_name = global_secondary_indexes[gsi_name.to_sym][:hash_key]
          find_all_by_index_and_hash_key(hash_key_name, hash_key_value, opts, gsi_name.to_s)
        end

        def find_all_by_gsi_hash_and_range_keys(gsi_name, hash_key_value, range_key_value)
          gsi_config = global_secondary_indexes[gsi_name.to_sym]
          find_all_by_index_hash_and_range_keys(
            hash_config: { name: gsi_config[:hash_key], value: hash_key_value },
            range_config: { name: gsi_config[:range_key], value: range_key_value },
            index_name: gsi_name.to_s
          )
        end

        def find_all_by_lsi_hash_key(lsi_name, hash_key_value, opts = {})
          hash_key_name = local_secondary_indexes[lsi_name.to_sym][:hash_key]
          find_all_by_index_and_hash_key(hash_key_name, hash_key_value, opts, lsi_name.to_s)
        end

        def find_all_by_lsi_hash_and_range_keys(lsi_name, hash_key_value, range_key_value)
          lsi_config = local_secondary_indexes[lsi_name.to_sym]
          find_all_by_index_hash_and_range_keys(
            hash_config: { name: lsi_config[:hash_key], value: hash_key_value },
            range_config: { name: lsi_config[:range_key], value: range_key_value },
            index_name: lsi_name.to_s
          )
        end

        def find_all_by_index_and_hash_key(hash_key_name, hash_key_value, opts = {}, index_name = nil)
          query_options = {
            select: 'ALL_ATTRIBUTES',
            key_condition_expression: "#{hash_key_name} = :hash_key_value",
            expression_attribute_values: {
              ':hash_key_value': hash_key_value
            },
            scan_index_forward: true
          }
          query_options[:index_name] = index_name if index_name
          query_options.merge!(opts)
          query(query_options)
        end

        def find_all_by_index_hash_and_range_keys(hash_config:, range_config:, index_name: nil,
          scan_index_forward: true, limit: nil)
          range_expression = range_config[:expression] || "#{range_config[:name]} = :rkv"
          query_options = {
            select: 'ALL_ATTRIBUTES',
            key_condition_expression: "#{hash_config[:name]} = :hkv AND #{range_expression}",
            expression_attribute_values: {
              ':hkv': hash_config[:value],
              ':rkv': range_config[:value]
            },
            scan_index_forward: scan_index_forward,
            limit: limit
          }
          query_options[:index_name] = index_name if index_name
          query(query_options)
        end

        def composite_key(*args)
          args.join(COMPOSITE_DELIMITER)
        end

        def split_composite(string)
          string.split(COMPOSITE_DELIMITER)
        end

        def max_retries
          DEFAULT_MAX_RETRIES
        end
      end

      module InstanceMethods
        def read_attribute_for_serialization(attribute)
          send(attribute)
        end

        def attribute_hash
          attrs = self.class.attributes.attributes
          attr_keys = attrs.keys
          hash = {}

          attr_keys.each do |key|
            hash[key] = attrs[key].type_cast(send(key))
          end
          hash
        end
      end
    end
  end
end
