# frozen_string_literal: true

module Dynamo
  module Record
    module TaskHelpers
      class Scale
        attr_reader :model, :attribute_selector, :new_throughput, :migration, :existing_throughput, :model_name

        def initialize(model_name, attribute_selector, new_throughput)
          @model_name = model_name
          @attribute_selector = attribute_selector
          @new_throughput = new_throughput
        end

        def run
          return description if [model_name, attribute_selector, new_throughput].any?(&:nil?)

          @model = model_name.constantize
          @migration = Aws::Record::TableMigration.new(model)
          @existing_throughput = model.provisioned_throughput

          update_throughput
          success_message
        end

        private

        def success_message
          "Successfully updated #{model.table_name} throughput to #{update_instructions[:provisioned_throughput]}"
        end

        def update_throughput
          raise_attribute_error if !read? && !write?

          migration.update!(update_instructions)
        end

        def update_instructions
          {
            provisioned_throughput: {
              write_capacity_units: (write? && new_throughput) || existing_write,
              read_capacity_units: (read? && new_throughput) || existing_read
            }
          }
        end

        def existing_write
          existing_throughput[:write_capacity_units]
        end

        def existing_read
          existing_throughput[:read_capacity_units]
        end

        def both?
          attribute_selector.to_sym == :both
        end

        def read?
          both? || attribute_selector.to_sym == :read
        end

        def write?
          both? || attribute_selector.to_sym == :write
        end

        def raise_attribute_error
          raise ArgumentError, <<~MESSAGE
            You didn't provide an appropriate attribute selection. We accept [:both, :read, :write]
          MESSAGE
        end

        def description
          <<~DESCRIPTION
            ----------------------------------------------------------------------
            Here's some usage information:
              Scale a dynamo table.  Requires three inputs.
                - ModelName
                  ruby class of the model
                - attribute
                  valid values include "both", "read" and "write"
                - new_throughput
                  numerical value for the new read/write capacity units
              Example: `rake dynamo:scale[MySuperDynamoModel,both,50]`
            ----------------------------------------------------------------------
          DESCRIPTION
        end
      end
    end
  end
end
